// Sound Blaster 2.0 / Pro / 16 (or compatibles) driver for PT4DOS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include "sb.h"
#include "../audio.h"
#include "../../asm.h"
#include "../../hardware/dma.h"
#include "../../hardware/irq.h"

#define dspTimeConstant		0x40
#define dspBlockSize		0x48
#define dspTurnSpeakerON	0xD1
#define dspTurnSpeakerOFF	0xD3
#define dspGetDSPVersion	0xE1
#define dspMixerAddr		0x04
#define dspMixerData		0x05
#define dspReset		0x06
#define dspReadData		0x0A
#define dspWriteBuf		0x0C
#define dspDataAvail		0x0E
#define dspDMA8Ack		0x0E
#define dspDMA16Ack		0x0F
#define dspReady		0xAA
#define mixVocVol		0x04
#define mixMicVol		0x0A
#define mixOutFilter		0x0E

static bool SBInitialized;
static uint8_t timeConstant, dmaNum, dma8 = 255, dma16 = 255, sbType;
static uint8_t oldMicVol, oldVoiceVol, oldFilterStereo;
static uint16_t dspVer, port;

// globals
uint8_t irqNum;

// asb.asm
extern uint16_t dmaAckPort;
extern void __interrupt sbLoIrqInterrupt(void);
extern void __interrupt sbHiIrqInterrupt(void);

static void dspOut(uint8_t val)
{
	while (inp(port+dspWriteBuf) & 0x80);
	outp(port+dspWriteBuf, val);
}

static void dspIn(uint8_t *val)
{
	while (!(inp(port+dspDataAvail) & 0x80));
	*val = inp(port+dspReadData);
}

static bool readEnvVars(void) // from Sound Blaster SDK (I think?)
{
	char buffer[5], *envString, charSave;
	int16_t digit, i, multiplier;
	bool dmaFound = false, portFound = false, irqFound = false;

	envString = getenv("BLASTER");
	if (envString == NULL)
		return false;

	do
	{
		switch (*envString)
		{
			case 'A':  // I/O base port address found
			case 'a':
				envString++;
				for (i = 0; i < 3; i++) // grab the digits
				{
					buffer[i] = *envString;
					envString++;
				}

				// the string is in hex, convert it to decimal
				multiplier = 1;
				port = 0;
				for (i -= 1; i >= 0; i--)
				{
					if (buffer[i] >= '0' && buffer[i] <= '9')
						digit = buffer[i] - '0';
					else if (buffer[i] >= 'A' && buffer[i] <= 'F')
						digit = buffer[i] - 'A' + 10;
					else if (buffer[i] >= 'a' && buffer[i] <= 'f')
						digit = buffer[i] - 'a' + 10;

					port = port + digit * multiplier;
					multiplier <<= 4;
				}

				portFound = true;
			break;

			case 'D': // 8-bit DMA channel
			case 'd':
			{
				charSave = *envString;
				envString++;
				buffer[0] = *envString;
				envString++;

				if (*envString >= '0' && *envString <= '9')
				{
					buffer[1] = *envString; // DMA channel is 2 digits
					buffer[2] = 0;
					envString++;
				}
				else
				{
					buffer[1] = 0; // DMA channel is 1 digit
				}

				dma8 = atoi(buffer); // 8-Bit DMA channel
				dmaFound = true;
			}
			break;
			
			case 'H': // 16-bit DMA channel
			case 'h':
			{
				charSave = *envString;
				envString++;
				buffer[0] = *envString;
				envString++;

				if (*envString >= '0' && *envString <= '9')
				{
					buffer[1] = *envString; // DMA channel is 2 digits
					buffer[2] = 0;
					envString++;
				}
				else
				{
					buffer[1] = 0; // DMA channel is 1 digit
				}

				dma16 = atoi(buffer); // 16-bit DMA channel
				dmaFound = true;
			}
			break;

			case 'I': // IRQ number
			case 'i':
			{
				envString++;
				buffer[0] = *envString;
				envString++;

				if (*envString >= '0' && *envString <= '9')
				{
					buffer[1] = *envString; // IRQ is 2 digits
					buffer[2] = 0;
					envString++;
				}
				else
				{
					buffer[1] = 0; // IRQ is 1 digit
				}

				irqNum = atoi(buffer);
				irqFound = true;
			}
			break;

			default: envString++; break;
		}
	}
	while (*envString != NULL);

	if (!dmaFound || !portFound || !irqFound)
		return false;

	return true;
}

static void SB_Reset(void)
{
	__asm
	{
		push eax
		push ecx
		push edx

		cli

		mov dx,port
		add dx,dspReset
		mov al,1
		out dx,al

		// wait at least 3 microseconds
		mov ecx,16
	L0:	in al,dx
		loop L0

		xor al,al
		out dx,al

		sti

		pop edx
		pop ecx
		pop eax
	}
}

bool SB_Detect(void)
{
	uint8_t val, hi, lo;
	bool detected = false;
	
	sbType = SBUnknown;

	port = 0;

	if (!readEnvVars())
		return false;

	if (port < 0x220 || port > 0x300 ||
		dma8 == 0 || (dma8 != 255 && dma8 >= 4) ||
		dma16 == 0 || (dma16 != 255 && dma16 > 7) ||
		irqNum < 1 || irqNum > 15)
	{
		return false;
	}

	SB_Reset();
	
	_disable();
	dspIn(&val);
	if (val == dspReady)
	{
		dspOut(dspGetDSPVersion);
		dspIn(&hi);
		dspIn(&lo);
		
		dspVer = (hi << 8) | lo;
		
		// we need an SB that supports >31388Hz for PT4DOS
		if (dspVer >= 0x201)
		{
			if (dspVer >= 0x400)
				sbType = SB16;
			else if (dspVer >= 0x300)
				sbType = SBPro;
			else if (dspVer >= 0x201)
				sbType = SB2;
			
			detected = true;
		}
	}
	_enable();

	return detected;
}

static void SetMixer(void)
{
	_disable();

	// read old volumes
	outp(port+dspMixerAddr, mixMicVol);
	oldMicVol = inp(port+dspMixerData);
	outp(port+dspMixerAddr, mixVocVol);
	oldVoiceVol = inp(port+dspMixerData);

	// set mic volume to zero
	outp(port+dspMixerAddr, mixMicVol);
	outp(port+dspMixerData, 0x00);

	// set max voice volume
	outp(port+dspMixerAddr, mixVocVol);
	outp(port+dspMixerData, 0xFF);

	if (sbType != SB16)
	{
		// read old stereo/filter mode
		outp(port+dspMixerAddr, mixOutFilter);
		oldFilterStereo = inp(port+dspMixerData);

		// disable output filter
		outp(port+dspMixerAddr, mixOutFilter);
		outp(port+dspMixerData, 0x20);
	}

	_enable();
}

static void RestoreMixer(void)
{
	_disable();

	// set back old mic volume
	outp(port+dspMixerAddr, mixMicVol);
	outp(port+dspMixerData, oldMicVol);

	// set back old voice volume
	outp(port+dspMixerAddr, mixVocVol);
	outp(port+dspMixerData, oldVoiceVol);

	if (sbType != SB16)
	{
		// set back old stereo/filter mode
		outp(port+dspMixerAddr, mixOutFilter);
		outp(port+dspMixerData, oldFilterStereo);
	}

	_enable();
}

bool SB_StartDMA(void)
{
	dspOut(dspTurnSpeakerON);

	if (!DMA_Start(dmaNum, DMA_PAUSE_IRQS))
		return false;

	_disable();
	if (sbType == SB16)
	{
		dspOut(dspTimeConstant);
		dspOut(timeConstant);

		dspOut(0xB6); // 16-bit output
		dspOut(0x10); // 16-bit mono signed PCM

		// block size (highest possible for less IRQ pollution)
		dspOut(0xFF);
		dspOut(0xFF);
	}
	else // SB 2.0 / SB Pro / SB Pro 2.0
	{
		dspOut(dspTimeConstant);
		dspOut(timeConstant);

		// block size (highest possible for less IRQ pollution)
		dspOut(dspBlockSize);
		dspOut(0xFF);
		dspOut(0xFF);

		dspOut(0x90); // auto-init 8-bit PCM high-speed output
	}
	_enable();

	return true;
}

bool SB_Init(void) // must be ran after SB_Detect()
{
	SBInitialized = false;

	if (sbType == SB2)
	{
		timeConstant = 232; // ~41666.67Hz @ mono (max on SB 2.0, unfortunately)
		output16Bit = false;
		
		dmaNum = dma8;
	}
	else if (sbType == SBPro)
	{
		timeConstant = 233; // ~43478.26Hz @ mono (234 possible, but not on (some?) clones)
		output16Bit = false;
		
		dmaNum = dma8;
	}
	else // SB16
	{
		timeConstant = 234; // ~45454.55Hz @ mono
		output16Bit = true;
		
		if (dma16 == 255)
			dma16 = dma8;
			
		dmaNum = dma16;
	}

	if (!DMA_Init(dmaNum, AUDIO_BUF_SIZE, DMA_MODE_READ | DMA_MODE_AUTOINIT | DMA_MODE_SINGLE))
	{
		printf("ERROR: Couldn't initialize audio DMA!\n");
		return false;
	}

	audio.dmaBuffer = DMA_GetBuffer(dmaNum);
	if (output16Bit)
		ASM_FastFill32(audio.dmaBuffer, 0x00000000, AUDIO_BUF_SIZE/4); // signed 16-bit
	else
		ASM_FastFill32(audio.dmaBuffer, 0x80808080, AUDIO_BUF_SIZE/4); // unsigned 8-bit

	audio.dOutputFreq = 1000000.0 / (double)(256 - timeConstant);

	if (sbType == SBPro || sbType == SB16)
		SetMixer();

	// for asb.asm
	dmaAckPort = output16Bit ? (port+dspDMA16Ack) : (port+dspDMA8Ack); 

	// for apitint.asm
	audio.sampleShift = output16Bit ? 1 : 0;
	audio.dmaNum = dmaNum;
	audio.dmaCountPort = dmaPorts[dmaNum].count;
	audio.dmaFlipflopPort = dmaPorts[dmaNum].flipflop;

	IRQ_SetIntVector(irqNum, (irqNum < 8) ? sbLoIrqInterrupt : sbHiIrqInterrupt);
	IRQ_Enable(irqNum);

	SBInitialized = true;
	return true;
}

void SB_Close(void)
{
	if (!SBInitialized)
		return;

	if (sbType == SBPro || sbType == SB16)
		RestoreMixer();

	IRQ_RestoreIntVector(irqNum);
	DMA_Close(dmaNum);

	SB_Reset();

	if (port != 0)
		dspOut(dspTurnSpeakerOFF);
}

uint8_t SB_GetModel(void)
{
	return sbType;
}
