#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "hardware/mouse.h"
#include "hardware/video.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "dir.h"
#include "ui.h"
#include "pattern.h"
#include "sampler.h"
#include "diskop.h"

#define FILELIST_X 8
#define FILELIST_Y 35

#define VISIBLE_FILE_ENTRIES 10

static char txtBuf[32];
static bool editOpWasShown, refreshDirectory = true;
static int32_t fileOffset;

static void DrawPath(void)
{
	Video_SelectWritePlane(2);
	
	Font_Print(24, 25, "___________________________________");
	if (Dir_CurPath == NULL)
		return;

	Font_Print(24, 25, Dir_CurPath);
}

void DiskOp_Refresh(void)
{
	UI_SetStatusText("READING DIR...", NO_CARRY);
	Mouse_SetDiskAccessColor();
		
	Dir_Read();
	fileOffset = 0;
	DiskOp_DrawFileList();
	
	Mouse_UpdateColor();
	UI_SetStatusText("SELECT MODULE", DO_CARRY);
}

bool DiskOp_SetDrive(char letter)
{
	if (Dir_SetDrive(letter))
	{
		Dir_GetPath();
		DrawPath();
		DiskOp_Refresh();
		return true;
	}

	return false;
}

bool DiskOp_SetPath(const char *path)
{
	if (Dir_SetPath(path))
	{
		Dir_GetPath();
		DrawPath();
		DiskOp_Refresh();
		return true;
	}

	return false;
}

bool DiskOp_HandleListSelection(void)
{
	int32_t entry;
	dir_t *d;

	if (Dir_NumFiles == 0 || mouse.y < FILELIST_Y || mouse.y >= FILELIST_Y+(VISIBLE_FILE_ENTRIES*(FONT_H+1)))
		return false;

	if (mouse.x < FILELIST_X || mouse.x > 302)
		return false;

	entry = fileOffset + ((mouse.y - FILELIST_Y) / (FONT_H+1));
	if (entry < 0 || entry >= Dir_NumFiles)
		return false;

	Mouse_SetDiskAccessColor();

	d = &Dir_List[entry];
	if (d->isDrive)
	{
		if (!DiskOp_SetDrive(d->name[1])) // [A], [1] = 'A'
			UI_Error("DRIVE ERROR !");
	}
	else if (d->isDir)
	{
		if (!DiskOp_SetPath(d->name))
			UI_Error("DIR ERROR !");
	}
	else
	{
		UI_SetStatusText("LOADING...", NO_CARRY);

		if (!Replayer_Load(d->name))
		{
			switch (Replayer_GetLoadError())
			{
				case LERR_FILE:
					UI_Error("CAN'T OPEN FILE !");
				break;

				case LERR_IO:
					UI_Error("GENERAL I/O ERROR");
				break;

				case LERR_OUT_OF_MEM:
					UI_Error("OUT OF MEMORY !");
				break;

				case LERR_NOT_A_MOD:
					UI_Error("UNSUPPORTED MOD !");
				break;

				default: 
					UI_Error("UNKNOWN ERROR !");
				break;
			}
		}
		else
		{
			ui.currSample = 0;
			UI_DrawSongName();
			UI_DrawSampleName();

			Replayer_Stop();

			DiskOp_Exit();
			UI_DrawSampleInfo();
			Sampler_UpdateSampleInfo();

			Replayer_CalculateModSize();
			Pattern_ClearCache();
			ui.redrawPositionFlag = true;
			ui.redrawBPMFlag = true;
			ui.redrawTempoModeFlag = true;
			ui.redrawPatternFlag = true;
			ui.redrawModSizeFlag = true;
		}
		
		UI_SetPrevStatusMessage();
	}

	Mouse_UpdateColor();
	return true;
}

void DiskOp_ScrollUp(void)
{
	if (Dir_NumFiles <= VISIBLE_FILE_ENTRIES)
		return;

	if (fileOffset > 0)
	{
		fileOffset--;
		DiskOp_DrawFileList();
	}
}

void DiskOp_ScrollDown(void)
{
	if (Dir_NumFiles <= VISIBLE_FILE_ENTRIES)
		return;

	if (fileOffset < Dir_NumFiles-VISIBLE_FILE_ENTRIES)
	{
		fileOffset++;
		DiskOp_DrawFileList();
	}
}

void DiskOp_ScrollToTop(void)
{
	uint32_t oldFileOffset = fileOffset;

	fileOffset = 0;
	if (fileOffset != oldFileOffset)
		DiskOp_DrawFileList();
}

void DiskOp_ScrollToBottom(void)
{
	uint32_t oldFileOffset = fileOffset;

	if (Dir_NumFiles <= VISIBLE_FILE_ENTRIES)
		fileOffset = 0;
	else
		fileOffset = Dir_NumFiles-VISIBLE_FILE_ENTRIES;

	if (fileOffset != oldFileOffset)
		DiskOp_DrawFileList();
}

void DiskOp_DrawFileList(void)
{
	int32_t i, y, entriesToList;
	dir_t *d;

	Video_SelectWritePlane(2);

	if (Dir_NumFiles < VISIBLE_FILE_ENTRIES)
		GFX_ClearDiskOpFileList(FILELIST_Y+(Dir_NumFiles*(FONT_H+1)), VISIBLE_FILE_ENTRIES-Dir_NumFiles);
	
	if (Dir_NumFiles >= VISIBLE_FILE_ENTRIES && fileOffset > Dir_NumFiles-VISIBLE_FILE_ENTRIES)
		fileOffset = Dir_NumFiles-VISIBLE_FILE_ENTRIES;
		
	if (fileOffset < 0)
		fileOffset = 0;
	
	if (Dir_List == NULL || Dir_NumFiles == 0 || fileOffset >= Dir_NumFiles)
		return;

	d = &Dir_List[fileOffset];

	entriesToList = VISIBLE_FILE_ENTRIES;
	if (entriesToList > Dir_NumFiles)
		entriesToList = Dir_NumFiles;

	y = FILELIST_Y;
	for (i = 0; i < entriesToList; i++, y += FONT_H+1, d++)
	{
		Font_Print(FILELIST_X+( 0*8), y, d->date);
		Font_Print(FILELIST_X+(11*8), y, d->name);
		Font_Print(FILELIST_X+(30*8), y, d->filesize);
	}
}

void DiskOp_Toggle(void)
{
	ui.diskOpShown ^= 1;
	
	if (ui.diskOpShown)
		DiskOp_Open();
	else
		DiskOp_Exit();
}

void DiskOp_Open(void)
{
	editOpWasShown = ui.editOpShown;
	ui.editOpShown = false;

	GFX_DrawDiskOpBitmap();
	ui.diskOpShown = true;
	
	DrawPath();

	if (refreshDirectory)
	{
		UI_SetStatusText("READING DIR...", NO_CARRY);
		Mouse_SetDiskAccessColor();
		Dir_Read();
		refreshDirectory = false;
		Mouse_UpdateColor();
	}
	
	DiskOp_DrawFileList();
	UI_SetStatusText("SELECT MODULE", DO_CARRY);
}

void DiskOp_Exit(void)
{
	UI_DrawTopScreen();
	ui.diskOpShown = false;

	if (editOpWasShown)
	{
		// TODO: show edit op. screen
	}

	UI_StatusAllRight();
}
