#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include "irq.h"

#define IRQ_NUM 16

static bool irqUsed[IRQ_NUM], irqWasEnabled[IRQ_NUM];
static void __interrupt (*oldIntVector[IRQ_NUM])(void);

void IRQ_SendEOI(uint8_t irqNum)
{
	if (irqNum > 7)
		outp(PIC2MODE, PICEOI); // extended IRQ, send EOI to PIC chip #2 first

	outp(PIC1MODE, PICEOI); // send EOI to PIC chip #1
}

void IRQ_Enable(uint8_t irqNum)
{
	if (irqNum >= IRQ_NUM)
		return;

	if (irqNum > 7)
		outp(PIC2MASK, inp(PIC2MASK) & ~(1 << (irqNum & 7)));
	else
		outp(PIC1MASK, inp(PIC1MASK) & ~(1 << irqNum));
}

void IRQ_Disable(uint8_t irqNum)
{
	if (irqNum >= IRQ_NUM)
		return;

	if (irqNum > 7)
		outp(PIC2MASK, inp(PIC2MASK) | (1 << (irqNum & 7)));
	else
		outp(PIC1MASK, inp(PIC1MASK) | (1 << irqNum));
}

void (__interrupt *IRQ_GetIntVector(uint8_t irqNum))(void)
{
	union REGS r;
	struct SREGS sr;

	if (irqNum >= IRQ_NUM)
		return NULL;

	_disable();

	segread(&sr);
	r.w.ax = 0x3500 + ((irqNum > 7) ? (0x70 + (irqNum & 7)) : (0x08 + irqNum));
	int386x(0x21, &r, &r, &sr);

	_enable();

	return MK_FP(sr.es, r.w.bx);
}

bool IRQ_SetIntVector(uint8_t irqNum, void (__interrupt *newIntVect)(void))
{
	union REGS r;
	struct SREGS sr;
    
	if (irqNum >= IRQ_NUM || irqUsed[irqNum])
		return false;

	_disable();
	
	if (irqNum > 7)
		irqWasEnabled[irqNum] = ((inp(PIC2MASK) >> (1 << (irqNum & 7))) & 1) ^ 1;
	else
		irqWasEnabled[irqNum] = ((inp(PIC1MASK) >> (1 << irqNum)) & 1) ^ 1;

	IRQ_Disable(irqNum);
	oldIntVector[irqNum] = IRQ_GetIntVector(irqNum);

	segread(&sr);
	r.w.ax = 0x2500 + ((irqNum > 7) ? (0x70 + (irqNum & 7)) : (0x08 + irqNum));
	r.x.edx = FP_OFF(newIntVect);
	sr.ds = FP_SEG(newIntVect);
	int386x(0x21, &r, &r, &sr);

	irqUsed[irqNum] = true;

	_enable();

	return true;
}

bool IRQ_RestoreIntVector(uint8_t irqNum)
{
	union REGS r;
	struct SREGS sr;

	if (irqNum >= IRQ_NUM || !irqUsed[irqNum])
		return false;

	IRQ_Disable(irqNum);
	IRQ_SendEOI(irqNum);

	segread(&sr);
	r.w.ax = 0x2500;
	r.h.al = (irqNum > 7) ? (0x70 + (irqNum & 7)) : (0x08 + irqNum);
	r.x.edx = FP_OFF(oldIntVector[irqNum]);
	sr.ds = FP_SEG(oldIntVector[irqNum]);
	int386x(0x21, &r, &r, &sr);

	if (irqWasEnabled[irqNum])
		IRQ_Enable(irqNum);

	irqUsed[irqNum] = false;
	return true;
}
